@@ -36,6 +36,151 @@ from fcntl import lockf, LOCK_EX, LOCK_UN, LOCK_NB
 logger = logging.getLogger(__name__)
 
 
+def log_errors(fullpath, cmdargs, lines):
+    logger.critical('%s reports errors:', fullpath)
+    with open(os.path.join(fullpath, 'grokmirror.fsck.err'), 'w') as fh:
+        fh.write('# Date: %s\n' % datetime.datetime.today().strftime('%F'))
+        fh.write('# Cmd : git %s\n' % ' '.join(cmdargs))
+        count = 0
+        for line in lines:
+            fh.write('%s\n' % line)
+            logger.critical('\t%s', line)
+            count += 1
+            if count > 10:
+                logger.critical('\t [ %s more lines skipped ]', len(lines) - 10)
+                logger.critical('\t [ see %s/grokmirror.fsck.err ]', os.path.basename(fullpath))
+                break
+
+
+def get_blob_set(fullpath):
+    bset = set()
+    size = 0
+    blobcache = os.path.join(fullpath, 'grokmirror.blobs')
+    if os.path.exists(blobcache):
+        # Did it age out? Hardcode to 30 days.
+        expage = time.time() - 86400*30
+        st = os.stat(blobcache)
+        if st.st_mtime < expage:
+            os.unlink(blobcache)
+    try:
+        with open(blobcache) as fh:
+            while True:
+                line = fh.readline()
+                if not len(line):
+                    break
+                if line[0] == '#':
+                    continue
+                chunks = line.strip().split()
+                bhash = chunks[0]
+                bsize = int(chunks[1])
+                size += bsize
+                bset.add((bhash, bsize))
+        return bset, size
+    except FileNotFoundError:
+        pass
+
+    # This only makes sense for repos not using alternates, so make sure you check first
+    logger.info(' bloblist: %s', fullpath)
+    gitargs = ['cat-file', '--batch-all-objects', '--batch-check', '--unordered']
+    retcode, output, error = grokmirror.run_git_command(fullpath, gitargs)
+    if retcode == 0:
+        with open(blobcache, 'w') as fh:
+            fh.write('# Blobs and sizes used for sibling calculation\n')
+            for line in output.split('\n'):
+                if line.find(' blob ') < 0:
+                    continue
+                chunks = line.strip().split()
+                fh.write(f'{chunks[0]} {chunks[2]}\n')
+                bhash = chunks[0]
+                bsize = int(chunks[2])
+                size += bsize
+                bset.add((bhash, bsize))
+
+    return bset, size
+
+
+def check_sibling_repos_by_blobs(bset1, bsize1, bset2, bsize2, ratio):
+    iset = bset1.intersection(bset2)
+    if not len(iset):
+        return False
+    isize = 0
+    for bhash, bsize in iset:
+        isize += bsize
+    # Both repos should share at least ratio % of blobs in them
+    ratio1 = int(isize / bsize1 * 100)
+    logger.debug('isize=%s, bsize1=%s, ratio1=%s', isize, bsize1, ratio1)
+    ratio2 = int(isize / bsize2 * 100)
+    logger.debug('isize=%s, bsize2=%s ratio2=%s', isize, bsize2, ratio1)
+    if ratio1 >= ratio and ratio2 >= ratio:
+        return True
+
+    return False
+
+
+def find_siblings_by_blobs(obstrepo, obstdir, ratio=75):
+    siblings = set()
+    oset, osize = get_blob_set(obstrepo)
+    for srepo in grokmirror.find_all_gitdirs(obstdir, normalize=True, exclude_objstore=False):
+        if srepo == obstrepo:
+            continue
+        logger.debug('Comparing blobs between %s and %s', obstrepo, srepo)
+        sset, ssize = get_blob_set(srepo)
+        if check_sibling_repos_by_blobs(oset, osize, sset, ssize, ratio):
+            logger.info(' siblings: %s and %s', os.path.basename(obstrepo), os.path.basename(srepo))
+            siblings.add(srepo)
+
+    return siblings
+
+
+def merge_siblings(siblings, amap):
+    mdest = None
+    rcount = 0
+    # Who has the most remotes?
+    for sibling in set(siblings):
+        if sibling not in amap or not len(amap[sibling]):
+            # Orphaned sibling, ignore it -- it will get cleaned up
+            siblings.remove(sibling)
+            continue
+        s_remotes = grokmirror.list_repo_remotes(sibling)
+        if len(s_remotes) > rcount:
+            mdest = sibling
+            rcount = len(s_remotes)
+
+    # Migrate all siblings into the repo with most remotes
+    siblings.remove(mdest)
+    for sibling in siblings:
+        logger.info('%s: merging into %s', os.path.basename(sibling), os.path.basename(mdest))
+        s_remotes = grokmirror.list_repo_remotes(sibling, withurl=True)
+        for virtref, childpath in s_remotes:
+            if childpath not in amap[sibling]:
+                # The child repo isn't even using us
+                args = ['remote', 'remove', virtref]
+                grokmirror.run_git_command(sibling, args)
+                continue
+            logger.info('   moving: %s', childpath)
+
+            success = grokmirror.add_repo_to_objstore(mdest, childpath)
+            if not success:
+                logger.critical('Could not add %s to %s', childpath, mdest)
+                continue
+
+            logger.info('         : fetching into %s', os.path.basename(mdest))
+            success = grokmirror.fetch_objstore_repo(mdest, childpath)
+            if not success:
+                logger.critical('Failed to fetch %s from %s to %s', childpath, os.path.basename(sibling),
+                                os.path.basename(mdest))
+                continue
+            logger.info('         : repointing alternates')
+            grokmirror.set_altrepo(childpath, mdest)
+            amap[sibling].remove(childpath)
+            amap[mdest].add(childpath)
+            args = ['remote', 'remove', virtref]
+            grokmirror.run_git_command(sibling, args)
+            logger.info('         : done')
+
+    return mdest
+
+
 def check_reclone_error(fullpath, config, errors):
     reclone = None
     toplevel = os.path.realpath(config['core'].get('toplevel'))
@@ -124,10 +269,8 @@ def run_git_prune(fullpath, config):
         if debug:
             logger.debug('Stderr: %s', '\n'.join(debug))
         if warn:
-            logger.critical('Pruning %s returned critical errors:', fullpath)
             prune_ok = False
-            for entry in warn:
-                logger.critical("\t%s", entry)
+            log_errors(fullpath, args, warn)
             check_reclone_error(fullpath, config, warn)
 
     if isprecious:
@@ -267,10 +410,8 @@ def run_git_repack(fullpath, config, level=1, prune=True):
         if debug:
             logger.debug('Stderr: %s', '\n'.join(debug))
         if warn:
-            logger.critical('Repacking %s returned critical errors:', fullpath)
             repack_ok = False
-            for entry in warn:
-                logger.critical("\t%s", entry)
+            log_errors(fullpath, args, warn)
             check_reclone_error(fullpath, config, warn)
 
     if not repack_ok:
@@ -311,11 +452,8 @@ def run_git_repack(fullpath, config, level=1, prune=True):
         if debug:
             logger.debug('Stderr: %s', '\n'.join(debug))
         if warn:
-            logger.critical('Repacking refs %s returned critical errors:', fullpath)
             repack_ok = False
-            for entry in warn:
-                logger.critical("\t%s", entry)
-
+            log_errors(fullpath, args, warn)
             check_reclone_error(fullpath, config, warn)
 
     if prune:
@@ -363,9 +501,7 @@ def run_git_fsck(fullpath, config, conn_only=False):
         if debug:
             logger.debug('Stderr: %s', '\n'.join(debug))
         if warn:
-            logger.critical('%s has critical errors:', fullpath)
-            for entry in warn:
-                logger.critical("\t%s", entry)
+            log_errors(fullpath, args, warn)
             check_reclone_error(fullpath, config, warn)
 
 
@@ -576,8 +712,14 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
     fetched_obstrepos = set()
     obst_changes = False
     analyzed = 0
+    queued = 0
     logger.info('Analyzing %s (%s repos)', toplevel, len(status))
+    stattime = time.time()
     for fullpath in list(status):
+        # Give me a status every 5 seconds
+        if time.time() - stattime >= 5:
+            logger.info('      ---: %s/%s analyzed, %s queued', analyzed, len(status), queued)
+            stattime = time.time()
         start_size = get_repo_size(fullpath)
         analyzed += 1
         # We do obstrepos separately below, as logic is different
@@ -614,7 +756,7 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
         else:
             m_prune = True
 
-        if not altdir:
+        if not altdir and not os.path.exists(os.path.join(fullpath, 'grokmirror.do-not-objstore')):
             # Do we match any obstdir repos?
             obstrepo = grokmirror.find_best_obstrepo(fullpath, obst_roots)
             if obstrepo:
@@ -639,7 +781,7 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
                     # Am I a private repo?
                     if is_private:
                         # Are there any non-private siblings?
-                        for top_sibling in grokmirror.find_siblings(fullpath, my_roots, top_roots):
+                        for top_sibling in top_siblings:
                             # Are you a private repo?
                             if grokmirror.is_private_repo(config, top_sibling):
                                 continue
@@ -672,7 +814,6 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
                         grokmirror.fetch_objstore_repo(obstrepo, fullpath)
                     run_git_repack(fullpath, config, level=1, prune=m_prune)
                     space_saved += start_size - get_repo_size(fullpath)
-                    logger.info('      ---: %s analyzed, %s queued, %s total', analyzed, len(to_process), len(status))
                     obst_roots[obstrepo] = grokmirror.get_repo_roots(obstrepo, force=True)
 
         elif not os.path.isdir(altdir):
@@ -698,8 +839,6 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
                         pre_size = get_repo_size(altdir)
                         run_git_repack(altdir, config, level=1, prune=False)
                         space_saved += pre_size - get_repo_size(altdir)
-                        logger.info('      ---: %s analyzed, %s queued, %s total', analyzed, len(to_process),
-                                    len(status))
                     else:
                         logger.critical('Unsuccessful fetching %s into %s', altdir, os.path.basename(obstrepo))
                         obstrepo = None
@@ -721,26 +860,31 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
                     pre_size = get_repo_size(altdir)
                     run_git_repack(altdir, config, level=1, prune=False)
                     space_saved += pre_size - get_repo_size(altdir)
-                    logger.info('      ---: %s analyzed, %s queued, %s total', analyzed, len(to_process), len(status))
                 else:
                     logger.critical('Unsuccessful fetching %s into %s', altdir, os.path.basename(obstrepo))
                     obstrepo = None
 
             if obstrepo:
                 obst_changes = True
-                # It should be safe now to repoint the alternates without doing a repack first
-                grokmirror.set_altrepo(fullpath, obstrepo)
                 if not is_private:
                     # Fetch into the obstrepo
                     grokmirror.add_repo_to_objstore(obstrepo, fullpath)
                     logger.info('    fetch: fetching %s', gitdir)
-                    grokmirror.fetch_objstore_repo(obstrepo, fullpath)
-                    set_precious_objects(fullpath, enabled=False)
-                    run_git_repack(fullpath, config, level=1, prune=m_prune)
-                    space_saved += start_size - get_repo_size(fullpath)
-                    logger.info('      ---: %s analyzed, %s queued, %s total', analyzed, len(to_process), len(status))
+                    if grokmirror.fetch_objstore_repo(obstrepo, fullpath):
+                        grokmirror.set_altrepo(fullpath, obstrepo)
+                        set_precious_objects(fullpath, enabled=False)
+                        run_git_repack(fullpath, config, level=1, prune=m_prune)
+                        space_saved += start_size - get_repo_size(fullpath)
                 else:
-                    logger.info('    fetch: not fetching %s (private)', gitdir)
+                    # Grab all the objects from the previous parent, since we can't simply
+                    # fetch ourselves into the obstrepo (we're private).
+                    args = ['repack', '-a']
+                    logger.info('    fetch: restoring private repo %s', gitdir)
+                    if grokmirror.run_git_command(fullpath, args):
+                        grokmirror.set_altrepo(fullpath, obstrepo)
+                        set_precious_objects(fullpath, enabled=False)
+                        # Now repack ourselves to get rid of any public objects
+                        run_git_repack(fullpath, config, level=1, prune=m_prune)
 
                 obst_roots[obstrepo] = grokmirror.get_repo_roots(obstrepo, force=True)
 
@@ -817,18 +961,20 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
                 repack_level = None
 
         if repack_level:
+            queued += 1
             to_process.add((fullpath, 'repack', repack_level))
             if repack_level > 1:
                 logger.info('   queued: %s (full repack)', fullpath)
             else:
                 logger.info('   queued: %s (repack)', fullpath)
-            logger.info('      ---: %s analyzed, %s queued, %s total', analyzed, len(to_process), len(status))
         elif repack_only or repack_all_quick or repack_all_full:
             continue
         elif schedcheck <= today or force:
+            queued += 1
             to_process.add((fullpath, 'fsck', None))
             logger.info('   queued: %s (fsck)', fullpath)
-            logger.info('      ---: %s analyzed, %s queued, %s total', analyzed, len(to_process), len(status))
+
+    logger.info('     done: %s analyzed, %s queued', analyzed, queued)
 
     if obst_changes:
         # Refresh the alt repo map cache
@@ -837,69 +983,42 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
         grokmirror.manifest_lock(manifile)
         manifest = grokmirror.read_manifest(manifile)
 
-    obstrepos = grokmirror.find_all_gitdirs(obstdir, normalize=True, exclude_objstore=False, flat=True)
+    obstrepos = grokmirror.find_all_gitdirs(obstdir, normalize=True, exclude_objstore=False)
 
     analyzed = 0
+    queued = 0
     logger.info('Analyzing %s (%s repos)', obstdir, len(obstrepos))
     baselines = [x.strip() for x in config['fsck'].get('baselines', '').split('\n')]
     islandcores = [x.strip() for x in config['fsck'].get('islandcores', '').split('\n')]
+    stattime = time.time()
     for obstrepo in obstrepos:
+        if time.time() - stattime >= 5:
+            logger.info('      ---: %s/%s analyzed, %s queued', analyzed, len(obstrepos), queued)
+            stattime = time.time()
         analyzed += 1
         logger.debug('Processing objstore repo: %s', os.path.basename(obstrepo))
         my_roots = grokmirror.get_repo_roots(obstrepo)
         if obstrepo in amap and len(amap[obstrepo]):
             # Is it redundant with any other objstore repos?
-            siblings = grokmirror.find_siblings(obstrepo, my_roots, obst_roots, exact=True)
+            strategy = config['fsck'].get('obstrepo_merge_strategy', 'exact')
+            if strategy == 'blobs':
+                siblings = find_siblings_by_blobs(obstrepo, obstdir, ratio=75)
+            else:
+                exact_merge = True
+                if strategy == 'loose':
+                    exact_merge = False
+                siblings = grokmirror.find_siblings(obstrepo, my_roots, obst_roots, exact=exact_merge)
             if len(siblings):
                 siblings.add(obstrepo)
-                mdest = None
-                rcount = 0
-                # Who has the most remotes?
-                for sibling in set(siblings):
-                    if sibling not in amap or not len(amap[sibling]):
-                        # Orphaned sibling, ignore it -- it will get cleaned up
-                        siblings.remove(sibling)
-                        continue
-                    s_remotes = grokmirror.list_repo_remotes(sibling)
-                    if len(s_remotes) > rcount:
-                        mdest = sibling
-                        rcount = len(s_remotes)
+                mdest = merge_siblings(siblings, amap)
+                obst_changes = True
+                if mdest in status:
+                    # Force full repack of merged obstrepos
+                    status[mdest]['nextcheck'] = todayiso
 
-                # Migrate all siblings into the repo with most remotes
-                siblings.remove(mdest)
-                for sibling in siblings:
-                    logger.info('%s: merging into %s', os.path.basename(sibling), os.path.basename(mdest))
-                    s_remotes = grokmirror.list_repo_remotes(sibling, withurl=True)
-                    for virtref, childpath in s_remotes:
-                        if childpath not in amap[sibling]:
-                            # The child repo isn't even using us
-                            args = ['remote', 'remove', virtref]
-                            grokmirror.run_git_command(sibling, args)
-                            continue
-                        logger.info('   moving: %s', childpath)
-
-                        success = grokmirror.add_repo_to_objstore(mdest, childpath)
-                        if not success:
-                            logger.critical('Could not add %s to %s', childpath, mdest)
-                            continue
-
-                        logger.info('         : fetching into %s', os.path.basename(mdest))
-                        success = grokmirror.fetch_objstore_repo(mdest, childpath)
-                        if not success:
-                            logger.critical('Failed to fetch %s from %s to %s', childpath, os.path.basename(sibling),
-                                            os.path.basename(mdest))
-                            continue
-                        logger.info('         : repointing alternates')
-                        grokmirror.set_altrepo(childpath, mdest)
-                        amap[sibling].remove(childpath)
-                        amap[mdest].add(childpath)
-                        args = ['remote', 'remove', virtref]
-                        grokmirror.run_git_command(sibling, args)
-                        logger.info('         : done')
-                        obst_changes = True
-                        if mdest in status:
-                            # Force full repack of merged obstrepos
-                            status[mdest]['nextcheck'] = todayiso
+                # Recalculate my roots
+                my_roots = grokmirror.get_repo_roots(obstrepo, force=True)
+                obst_roots[obstrepo] = my_roots
 
         # Not an else, because the previous step may have migrated things
         if obstrepo not in amap or not len(amap[obstrepo]):
@@ -1000,7 +1119,11 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
 
             manifest[gitdir]['forkgroup'] = os.path.basename(obstrepo[:-4])
 
-        if obstrepo not in status or new_islandcore:
+        repack_requested = False
+        if os.path.exists(os.path.join(obstrepo, 'grokmirror.repack')):
+            repack_requested = True
+
+        if obstrepo not in status or new_islandcore or repack_requested:
             # We don't use obstrepo fingerprints, so we set it to None
             status[obstrepo] = {
                 'lastcheck': 'never',
@@ -1019,19 +1142,21 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
             repack_level = 1
 
         if repack_level:
+            queued += 1
             to_process.add((obstrepo, 'repack', repack_level))
             if repack_level > 1:
                 logger.info('   queued: %s (full repack)', os.path.basename(obstrepo))
             else:
                 logger.info('   queued: %s (repack)', os.path.basename(obstrepo))
-            logger.info('      ---: %s analyzed, %s queued, %s total', analyzed, len(to_process), len(status))
         elif repack_only or repack_all_quick or repack_all_full:
             continue
         elif (nextcheck <= today or force) and not repack_only:
+            queued += 1
             status[obstrepo]['nextcheck'] = nextcheck.strftime('%F')
             to_process.add((obstrepo, 'fsck', None))
             logger.info('   queued: %s (fsck)', os.path.basename(obstrepo))
-            logger.info('      ---: %s analyzed, %s queued, %s total', analyzed, len(to_process), len(status))
+
+    logger.info('     done: %s analyzed, %s queued', analyzed, queued)
 
     if obst_changes:
         # We keep the same mtime, because the repos themselves haven't changed
@@ -1067,6 +1192,11 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
             if run_git_repack(fullpath, config, repack_level):
                 status[fullpath]['lastrepack'] = todayiso
                 if repack_level > 1:
+                    try:
+                        os.unlink(os.path.join(fullpath, 'grokmirror.repack'))
+                    except FileNotFoundError:
+                        pass
+
                     status[fullpath]['lastfullrepack'] = todayiso
                     status[fullpath]['lastcheck'] = todayiso
                     status[fullpath]['nextcheck'] = nextcheck.strftime('%F')
@@ -1099,7 +1229,11 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
             logger.info('     done: %ss, %s saved', elapsed, get_human_size(saved))
         else:
             logger.info('     done: %ss', elapsed)
-        logger.info('      ---: %s done, %s queued', total_checked, len(to_process)-total_checked)
+        if space_saved > 0:
+            logger.info('      ---: %s done, %s queued, %s saved', total_checked,
+                        len(to_process)-total_checked, get_human_size(space_saved))
+        else:
+            logger.info('      ---: %s done, %s queued', total_checked, len(to_process)-total_checked)
 
         # Write status file after each check, so if the process dies, we won't
         # have to recheck all the repos we've already checked
@@ -1107,7 +1241,7 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
         with open(statusfile, 'w') as stfh:
             stfh.write(json.dumps(status, indent=2))
 
-    logger.info('Processed %s repos in %0.2fs, %s saved', total_checked, total_elapsed, get_human_size(space_saved))
+    logger.info('Processed %s repos in %0.2fs', total_checked, total_elapsed)
 
     with open(statusfile, 'w') as stfh:
         stfh.write(json.dumps(status, indent=2))
